/**
 * \file Mini.hh
 * \brief Header file, containing the function headers for Mini.cpp
 * \author <a href="mailto:auto_pilot@gmx.at">Hoermann Klaus</a>
 * \author <a target="_blank" href="http://www.netmusic.at.tt">www.netmusic.at.tt</a>
 * \version 0.1
 * \date 24.03.2004
 * \date last modified 03.06.2005
 *
 */

/**
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#ifndef Mini_class
#define Mini_class

#include <gmpxx.h>
#include <cmath>

/**
 * Value defines the default number of digits used to hold the value k (
 * number of voices).
 */
const int DIGITS_K_DEF = 4;
/**
 * Value defines the default number of digits used to hold the value size (
 * how many bytes used to represent the chord-code).
 */
const int DIGITS_SIZE_DEF = 3;

/**
 * Provides methods for MINI decoding, encoding as well as the binary
 * operations.
 */
class Mini
{
	public:
		
		/**
		 * \brief function is used to encapsulate the process of mini encoding
		 * \brief NOT WORKING ! ! !
		 */
		static void set_mini_word( int* noteArray, int k, char* binaryMiniWord1,
			int noteOffValue, int* size_mem_mpz, int startKey);

		/**
		 * \brief encodes a integer array with k elements int a mpz_t value
		 * \param k Integer value holding the amount of elements (notes).
		 * \param a[] Integer array containing the different values (notes)
		 * \param rank Multiple precision type standing for the chord-code.
		 * \sa ksub_unrank()
		 */
		static void ksub_rank ( int k, int* a, mpz_t rank );

		/**
		 * \brief decodes a mpz_t value into a integer array with k elements.
		 * \param k Integer value holding the amount of elements (notes).
		 * \param a[] Integer array containing the different values (notes)
		 * \param rank Multiple precision type standing for the chord-code.
		 * \sa ksub_rank()
		 */
		static void ksub_unrank ( int k, int* a, mpz_t rank );

		/**
		 * \brief returns the size (in Byte) of memory used to store the int value
		 * \param value A integer value.
		 * \return Integer value how many bytes necessary to store value.
		 * \sa get_size_mem_mpz()
		 */
		static int get_size_mem_int ( int value );

		/**
		 * \brief calculates necessary digits to store int value in binary format.
		 * \param value A integer value.
		 * \return Amount of digits
		 * \sa get_bin_digits_mpz()
		 */
		static int get_bin_digits_int ( int value );

		/**
		 * \brief returns the size (in Byte) of memory used to store mpz_t value
		 * \param value A integer value.
		 * \return Integer value how many bytes necessary to store value.
		 * \sa get_size_mem_int()
		 */
		static int get_size_mem_mpz ( mpz_t value );

		/**
		 * \brief calculates necessary digits to store the mpz value in binary format.
		 * \param number A mpzt number (chord code).
		 * \return Amount of digits
		 * \sa get_bin_digits_int()
		 */
		static int get_bin_digits_mpz ( mpz_t number);

		/**
		 * \brief converts mpz value into a binary format
		 * \param value A mpz value containing the chord code.
		 * \param k A integer value containing the amount of voices.
		 * \param note the noteValue of the pressed key
		 * \param mem_pointer A memory pointer pointing to a certain address.
		 *
		 * The method converts a mpz value and the given parameters k, note 
		 * and the amount of bytes used to display the whole information, into
		 * a binary value.
		 *
		 * \sa int_2_bin()
		 */
		static void mpz_2_bin ( mpz_t value, int k, int note, char* mem_pointer );
		
		/**
		 * \brief converts the value at the location mem_pointer back to mpz_value
		 * \param mem_pointer Points to a char* in memory.
		 * \param k A integer value containing the amount of voices.
		 * \param value A mpz value containing the chord code.
		 * \param note noteOff value
		 *
		 * This method converts a given binary format back into k, the size of bytes
		 * used for the whole format, the mpz value and the noteOff value.
		 *
		 * \sa bin_2_int()
		 */
		static void bin_2_mpz ( char* mem_pointer, int* k, mpz_t value, int* note);
		
		/**
		 * \brief receives a char which holds the size of the mpz value.
		 * \param firstRead The byte which holds the value.
		 * \param size the size of the whole binary
		 */
		static void extract_size(char firstRead, int* size);

		/**
		 * \brief prints the binary content at a specific memory location
		 * \param mem_pointer The location from which the printing should start
		 * \param size_bin value containing the size (in Byte) of the printed memory
		 */
		static void print_mem ( char* mem_pointer, int size_bin );

		/**
		 * \brief sorts a array.
		 * \param noteArray A pointer to the memory where the array lies
		 * \param size The size of the array, how many entries
		 */
		static void sort_array( int* noteArray, int size );

		/**
		 * \brief prints out a integer array.
		 * \param noteArray A pointer to the memory where the array lies
		 * \param size The size of the array, how many entries
		 * \param startKey To print out the values correctly
		 */
		static void print_array( int* noteArray, int size, int startKey );
		
};

/**
 * \brief Inline function to calculate how many bytes are necessary to represent the int value binary.
 * \param value A integer value.
 * \sa get_size_mem_mpz()
 */
inline
int Mini::get_size_mem_int(int value)
{
	return (int)ceil((double)(1 + DIGITS_K_DEF + DIGITS_SIZE_DEF + get_bin_digits_int(value))/(double)(8));
}

/**
 * \brief Inline function to calculate how many bytes are necessary to represent the mpz value binary.
 * \param value A mpz value.
 * \sa get_size_mem_int()
 */
inline
int  Mini::get_size_mem_mpz(mpz_t value)
{
	return (int)ceil((double)(1 + DIGITS_K_DEF + DIGITS_SIZE_DEF + get_bin_digits_mpz(value))/(double)(8));
}

#endif

