/**
 * \file Netmusic.cpp
 * \brief Contains code for methods listed in Netmusic.hh 
 * \author <a href="mailto:auto_pilot@gmx.at">Hoermann Klaus</a>
 * \author <a target="_blank" href="http://www.netmusic.at.tt">www.netmusic.at.tt</a>
 * \version 0.1ss
 * \date 25.03.2005
 * \version last modified 31.05.2005
 *
 */

/**
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */
 
#include <gmpxx.h>

#include "Netmusic.hh"

//***************************************************************
void Netmusic::check_error( int value, std::string error_text)
{
	if(value < 0 )
	{
		std::cout << error_text << std::endl;
		_exit(0);
	}
}

//***************************************************************
snd_seq_t* Netmusic::sender_open_seq()
{
	snd_seq_t *seq_handle;
	int portid;

 	Netmusic::check_error(snd_seq_open(&seq_handle, "default", 
		SND_SEQ_OPEN_INPUT, 0), "Error opening sequencer sender");

	Netmusic::check_error(snd_seq_set_client_name(seq_handle,
		"Netmusic Sender"), "Error setting name sequencer sender");

	portid = snd_seq_create_simple_port(seq_handle, "Netmusic Sender",
            SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
            SND_SEQ_PORT_TYPE_APPLICATION);
	Netmusic::check_error(portid, "Error creating simple port sender");
	return(seq_handle);
}

//***************************************************************
snd_seq_t* Netmusic::receiver_open_seq(int* portId)
{
	snd_seq_t *seq_handle;

	Netmusic::check_error(snd_seq_open(&seq_handle, "default", 
		SND_SEQ_OPEN_OUTPUT, 0), "Error opening sequencer handle receiver");

	Netmusic::check_error(snd_seq_set_client_name(seq_handle,
		"Netmusic Receiver"), "Error setting name sequencer receiver");
	(*portId) = snd_seq_create_simple_port(seq_handle, "Netmusic Receiver",
            SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
            SND_SEQ_PORT_TYPE_APPLICATION);
	Netmusic::check_error((*portId), "Error creating simple port receiver");
	return(seq_handle);
}

//***************************************************************
void Netmusic::print_read_data(snd_seq_event_t* ev)
{
	// output read data
	switch(ev->type)
	{
		case SND_SEQ_EVENT_PGMCHANGE:
			std::cout << DELIMITER << std::endl;
			std::cout << "SENDER: MIDI Thread read Program Change: "
				  << (int)ev->data.control.value << std::endl;
			break;

		case SND_SEQ_EVENT_NOTEON:
			std::cout << DELIMITER << std::endl;
			// recognized as note on 'meaning' note OFF
			if(ev->data.note.velocity == 0)
				std::cout << "SENDER: MIDI Thread read Note Off: ";
			else
				std::cout << "SENDER: MIDI Thread read Note On:  ";
			std::cout << (int)ev->data.note.note << std::endl;
			break;

		case SND_SEQ_EVENT_NOTEOFF:
			std::cout << DELIMITER << std::endl;
			std::cout << "SENDER: MIDI Thread read Note Off:  ";
			std::cout << (int)ev->data.note.note << std::endl;
			break;
	}
}

//***************************************************************
void Netmusic::sender_print_welcome_msg()
{
	std::cout << std::endl << DELIMITER << std::endl;
	std::cout << "Netmusic " << VERSION << " - SENDER" << std::endl;
	std::cout << AUTHOR << " " << EMAIL << std::endl << std::endl;
}

//***************************************************************
void Netmusic::receiver_print_welcome_msg()
{
	std::cout << std::endl << DELIMITER << std::endl;
	std::cout << "Netmusic " << VERSION << " - RECEIVER" << std::endl;
	std::cout << AUTHOR << " " << EMAIL << std::endl << std::endl;
}

//***************************************************************
void Netmusic::sender_print_help_msg()
{
	std::cout << std::endl << DELIMITER << std::endl;
	std::cout << "Netmusic " << VERSION << " - 'Sender-HELP'" << std::endl << std::endl;
	std::cout << "-midi................... MIDI Protocol instead of MINI used." << std::endl;
	std::cout << "-portMidi=[x]........... Used port number for sending MIDI Messages." << std::endl;
	std::cout << "-portMini=[x]........... Used port number for sending MINI Messages." << std::endl;
	std::cout << "-portSync=[x]........... Used port number for synchronizing with the receiver." << std::endl;	
	std::cout << "-startKey=[x]........... The lowest key you want to use (default=36)." << std::endl;
	std::cout << "-receiverIp=[x]......... IP Address of the receiving computer." << std::endl;	
	std::cout << "-threshold=[x].......... Duration of chord window in miliseconds." << std::endl;
	std::cout << "-help................... Display this help text." << std::endl << std::endl;
}

//***************************************************************
void Netmusic::receiver_print_help_msg()
{
	std::cout << std::endl << DELIMITER << std::endl;
	std::cout << "Netmusic " << VERSION << " - 'Receiver-HELP'" << std::endl << std::endl;
	std::cout << "-portSync=[x]........... Used port number for synchronization." << std::endl;
	std::cout << "-pgmChange=[x].......... MIDI Instrument number for playback." << std::endl;
	std::cout << "-velocity=[x]........... MIDI velocity for playback." << std::endl;
	std::cout << "-help................... Display this help text." << std::endl << std::endl;
}

//***************************************************************
void Netmusic::sender_check_args(int argc, char** argv, 
	int* mini,
	int* midiPort,int* miniPort, int* syncPort,
	int* startKey, int* threshold, 
	char* receiverIp)
{
	for(int i=1;i<argc;i++)
	{
		// midi
		if(strncmp(argv[i], "-midi", 5)==0)
		{
			*mini = 0;
		}
		
		// midiPort
		else if(strncmp(argv[i], "-portMidi=", 10)==0)
		{
			strtok(argv[i], "=");
			*midiPort = atoi(strtok(NULL, " "));
			if((*midiPort) < 0)
				Netmusic::check_error(-1, "Error setting portMidi");
		}
		
		// miniPort
		else if(strncmp(argv[i], "-portMini=", 10)==0)
		{
			strtok(argv[i], "=");
			*miniPort = atoi(strtok(NULL, " "));
			if((*miniPort) < 0)
				Netmusic::check_error(-1, "Error setting portMini");
		}
		
		// miniPort
		else if(strncmp(argv[i], "-portSync=", 10)==0)
		{
			strtok(argv[i], "=");
			*syncPort = atoi(strtok(NULL, " "));
			if((*syncPort) < 0)
				Netmusic::check_error(-1, "Error setting portSync");
		}

		// startKey
		else if(strncmp(argv[i], "-startKey=", 10)==0)
		{
			strtok(argv[i], "=");
			*startKey = atoi(strtok(NULL, " "));
			if((*startKey)<0)
				Netmusic::check_error(-1, "Error setting startKey");
		}
		
		// receiverIp
		else if(strncmp(argv[i], "-receiverIp=", 12)==0)
		{
			strtok(argv[i], "=");
			char* tmp = strtok(NULL, " ");
			std::cout << &tmp << std::endl;
			strcpy(receiverIp, tmp);
		}

		// threshold
		else if(strncmp(argv[i], "-threshold=", 11)==0)
		{
			strtok(argv[i], "=");
			*threshold = atoi(strtok(NULL, " "));
			if((*threshold) < 0)
				Netmusic::check_error(-1, "Error setting threshold");
		}

		// help
		else if(strncmp(argv[i], "-help", 6)==0)
		{
			sender_print_help_msg();
			_exit(0);
		}
		
		// unrecognized option
		else
		{
			std::cout << "Unrecognized option: " << argv[i] << std::endl;
			sender_print_help_msg();
			_exit(0);
		}
	}
	
	// check if receiverIp is set 
	if(receiverIp == NULL)
	{
		std::cout << "You must set the -receiverIp." << std::endl;
		_exit(0);
	}
}

//***************************************************************
void Netmusic::receiver_check_args(int argc, char** argv, 
		int* syncPort, 
		int* pgmChange, int* velocity)
{
	
	for(int i=1;i<argc;i++)
	{
		// sync Port
		if(strncmp(argv[i], "-portSync=", 10)==0)
		{
			strtok(argv[i], "=");
			*syncPort = atoi(strtok(NULL, " "));
			if((*syncPort) < 0)
				Netmusic::check_error(-1, "Error setting portSync");
		}
		
		// help
		else if(strncmp(argv[i], "-help", 6)==0)
		{
			receiver_print_help_msg();
			_exit(0);
		}		
		
		// pgmChange option
		else if(strncmp(argv[i], "-pgmChange=", 11)==0)
		{
		  strtok(argv[i], "=");
		  *pgmChange = atoi(strtok(NULL, " "));
		  if((*pgmChange) < 0 || (*pgmChange) > 127)
		    Netmusic::check_error(-1, "Error setting pgmChange.");
		}

		// velocity
		else if(strncmp(argv[i], "-velocity=", 10)==0)
		{
			strtok(argv[i], "=");
			*velocity = atoi(strtok(NULL, " "));
			if((*velocity) < 0 || (*velocity) > 127)
				Netmusic::check_error(-1, "Error setting velocity.");
		}
		
		// unrecognized option
		else
		{
			std::cout << "Unrecognized option: " << argv[i] << std::endl;
			receiver_print_help_msg();
			_exit(0);
		}
	}
}

//***************************************************************
void Netmusic::sender_print_param(
	int mini,
	int midiPort,int miniPort, int syncPort,
	int startKey, int threshold, 
	char* receiverIp)
{
	std::cout << std::endl;
	
	// mini
	std::cout << "mini..................... ";
	if(mini==1)
		std::cout << "yes" << std::endl;
	else 
		std::cout << "no" << std::endl;
	// socket ports
	std::cout << "syncPort................. " << syncPort << std::endl;
	if(mini==1)
		std::cout << "miniPort................. " << miniPort << std::endl;
	else if(mini==0)
		std::cout << "midiPort................. " << midiPort << std::endl;
	// startKey
	if(mini==1)
		std::cout << "startKey................. " << startKey << std::endl;
	// receiverIp
	std::cout << "IP Address Receiver...... " << receiverIp << std::endl;
	// threshold
	if(mini==1)
		std::cout << "Threshold................ " << threshold << std::endl
			  << std::endl;
}

//***************************************************************
void Netmusic::receiver_print_param(
	int mini,
	int midiPort,int miniPort, int syncPort,
	int startKey, 
	int pgmChange, int velocity)
{	
	std::cout << std::endl;
	
	// mini
	std::cout << "mini..................... ";
	if(mini==1)
		std::cout << "yes" << std::endl;
	else 
		std::cout << "no" << std::endl;
	// socket ports
	std::cout << "syncPort................. " << syncPort << std::endl;
	if(mini==1)
		std::cout << "miniPort................. " << miniPort << std::endl;
	else if(mini==0)
		std::cout << "midiPort................. " << midiPort << std::endl;
	// pgmChange
	std::cout << "pgmChange instrument..... " << pgmChange << std::endl;
	// velocity
	std::cout << "velocity................. " << velocity << std::endl;
	// startKey
	if(mini==1)
		std::cout << "startKey................. " << startKey << std::endl
			  << std::endl;
}		
