/**
 * \file Netmusic.hh
 * \brief Header file containing the header functions for Netmusic.cpp
 * \author <a href="mailto:auto_pilot@gmx.at">Hoermann Klaus</a>
 * \author <a target="_blank" href="http://www.netmusic.at.tt">www.netmusic.at.tt</a>
 * \version 0.1
 * \date 25.03.2005
 * \date last modified 31.05.2005
 *
 */

/**
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#ifndef Netmusic_class
#define Netmusic_class

#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <unistd.h>

#include <alsa/asoundlib.h>

/**
 * Delimiter for output.
 */
const std::string DELIMITER = "------------------------------------------------";
/**
 * String holding the current version.
 */
const std::string VERSION = "0.1";
/**
 * String holding the author name.
 */
const std::string AUTHOR = "Klaus Hoermann";
/**
 * String holding the author email address.
 */
const std::string EMAIL = "<auto_pilot@gmx.at>";

/**
 * Provides methods for the netmusic sender and receiver program.
 */
class Netmusic
{
	public:
	
		/**
		 * \brief checks the return value of many other functions.
		 * \param value A integer value containing the to be checked value.
		 * \param error_text A string containing the error text 
		 */
		static void check_error(int value, std::string error_text);

		/**
		 * \brief creates a new sequencer s on the sender side
		 * \return seq_handle Returns the handler for the sequencer
		 * \sa receiver_open_seq()
		 */
		static snd_seq_t* sender_open_seq();
		
		/**
		 * \brief creates a new sequencer client on the receiver side
		 * \return  seq_handle Returns the handler for the sequencer
		 * \param portId The portId of the used output port.
		 * \sa sender_open_seq()
		 */
		static snd_seq_t* receiver_open_seq(int* portId);

		/**
		 * \brief write read data to console
		 * \param ev read event
		 */
		static void print_read_data(snd_seq_event_t* ev);

		/**
		 * \brief Prints out the welcome message for the sender
		 */
		static void sender_print_welcome_msg();

		/**
		 * \brief Prints out the welcome message for the receiver
		 */
		static void receiver_print_welcome_msg();

		/**
		 * \brief Prints out the help message text for the sender
		 */
		static void sender_print_help_msg();
		
		/**
		 * \brief Prints out the help message text for the receiver
		 */
		static void receiver_print_help_msg();

		/**
		 * \brief Prints out the set values of the application.
		 * \param mini A pointer to the mini parameter.
		 * \param midiPort
		 * \param miniPort
		 * \param syncPort
		 * \param startKey
		 * \param receiverIp
		 * \param threshold
		 */
		static void sender_print_param(
			int mini,
			int midiPort,int miniPort, int syncPort,
			int startKey, int threshold,
			char* receiverIp);

		/**
		 * \brief Prints out the receiver parameters
		 * \param mini
		 * \param midiPort
		 * \param miniPort
		 * \param syncPort
		 * \param startKey
		 * \param pgmChange
		 * \param velocity
		 */
		static void receiver_print_param(
			int mini,
			int midiPort,int miniPort, int syncPort,
			int startKey, int pgmChange, int velocity );

		/**
		 * \brief Sets the given sender parameters, according to the command line args.
		 * \param argc Number of args
		 * \param argv char** holds all args
		 * \param mini A pointer to the mini parameter.
		 * \param midiPort
		 * \param miniPort
		 * \param syncPort
		 * \param startKey
		 * \param receiverIp
		 * \param threshold
		 */
		static void sender_check_args(int argc, char** argv,
			int* mini,
			int* midiPort,int* miniPort, int* syncPort,
			int* startKey, int* threshold,
			char* receiverIp);
		
		/**
		 * \brief Sets the given receiver parameters, according to the command line args.
		 * \param argc
		 * \param argv
		 * \param syncPort
		 * \param pgmChange
		 * \param velocity
		 */
		static void receiver_check_args(int argc, char** argv,
			int* syncPort, 
			int* pgmChange, int* velocity);
};

#endif

