/**
 * \file netmusic_receiver.cpp
 * \brief Receiver program code
 * \author <a href="mailto:auto_pilot@gmx.at">Hoermann Klaus</a>
 * \author <a target="_blank" href="http://www.netmusic.at.tt">www.netmusic.at.tt</a>
 * \version 0.1
 * \date 23.03.2005
 * \date last modified 03.06.2005
 *
 */

/**
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <sys/time.h>
#include <netdb.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/msg.h>
#include <sys/ipc.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "include/Netmusic.hh"
#include "include/Mini.hh"

/**
 * ALSA Sound Sequencer handler
 */
snd_seq_t* receiverSeqHandler = NULL;
/**
 * ???
 */
int npfd;
/**
 *
 */
struct pollfd* pfd;
/**
 * Port id for the sequencer output.
 */
int portId;
/**
 * Parameter mini (gets synchronized)
 */
int mini;
/**
 * Parameter midi port (gets synchronized)
 */
int midiPort;
/**
 * Parameter mini port (gets synchronized)
 */
int miniPort;
/**
 * Parameter startKey (gets synchronized)
 */
int startKey;
/**
 * Parameter pgmChange instrument for playback
 */
int pgmChange = 1;
/**
 * Parameter velocity for the playback
 */
int velocity = 127;
/**
 * Parameter sync port (default 3000)
 */
int syncPort = 29998;
/**
 * Used threads: Mini, Midi, Sync
 */
pthread_t miniThread, midiThread, syncThread;


//********************************************************************
/**
 * \brief Signal handler for STRG-C
 * \param signal int value for the given signal
 */
void handler(int signal)
{
	if ( signal == SIGINT)
		_exit(0); // kills the process and all the threads
}

//********************************************************************
/*
 * \brief midiThread function
 * \param ptr empty pointer
 *
 * The midi_function is executed by the midi thread. This thread is generated
 * if mini==0. It receives messages from the sender, in form of
 * snd_seq_event_t values. They can be sent to the port directly.
 *
 */
void *midi_function(void *ptr)
{
	// welcome msg
  std::cout << "RECEIVER: MIDI Thread started..." << std::endl
  	<< std::endl << std::endl;
  
  // variables
  struct sockaddr_in senderAddr, receiverAddr;
  int sd, rc, n;
  snd_seq_event_t* ev = new snd_seq_event;
  
  // socket creation
	sd = socket(AF_INET, SOCK_DGRAM, 0);
	Netmusic::check_error(sd, "Error creating midi socket");
	// bind receiver
	receiverAddr.sin_family = AF_INET;
	receiverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	receiverAddr.sin_port = htons(midiPort);
	rc = bind(sd, (struct sockaddr *) &receiverAddr, sizeof(receiverAddr));
	Netmusic::check_error(rc, "Error binding midi to receiverAddr");
	
	// receive midi
	while(1)
	{
		// read note
		n = recvfrom(sd, ev, sizeof(snd_seq_event_t), 0,
			(struct sockaddr *) &senderAddr, (socklen_t *)sizeof(receiverAddr));
		
		std::cout << DELIMITER << std::endl;
		std::cout << "RECEIVER: Received ";
			
		if((*ev).type == SND_SEQ_EVENT_NOTEON )
		{
			if((*ev).data.note.velocity != 0)
				std::cout << "Note On:  " << (int)(*ev).data.note.note << std::endl;
			else
				std::cout << "Note Off: " << (int)(*ev).data.note.note << std::endl;
		}
		else if((*ev).type == SND_SEQ_EVENT_NOTEOFF )
			std::cout << "Note Off: " << (int)(*ev).data.note.note << std::endl;
		else if((*ev).type == SND_SEQ_EVENT_PGMCHANGE )
			std::cout << "Program Change: " <<  (int)(*ev).data.control.value << std::endl;
			
		// output event directly
		snd_seq_ev_set_subs(ev);
		snd_seq_ev_set_direct(ev);
		snd_seq_ev_set_source(ev, portId);
		snd_seq_event_output_direct(receiverSeqHandler, ev);
		snd_seq_free_event(ev);
	}
  pthread_exit(NULL);
}

//********************************************************************
/**
 * \brief miniThread function
 * \param ptr empty pointer
 *
 * This function is executed by the mini thread, which is started if
 * mini==1. This method waits for incoming MINI Messages. THey are then
 * decoded, and note on messages generated. 
 *
 */
void *mini_function(void *ptr)
{
	// welcome msg
  std::cout << "RECEIVER: MINI Thread..." << std::endl;
	
	// variables
	struct sockaddr_in senderAddr, receiverAddr;
	int sd, rc, n;
	int* noteArray;
	char* firstRead = new char;
	mpz_t chordCode;
	mpz_init(chordCode);
	char* secondRead;
	int k, noteOff, readSize;
	
	// socket creation
	sd = socket(AF_INET, SOCK_DGRAM, 0);
	Netmusic::check_error(sd, "Error creating mini socket");
	// bind receiver
	receiverAddr.sin_family = AF_INET;
	receiverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	receiverAddr.sin_port = htons(miniPort);
	rc = bind(sd, (struct sockaddr *) &receiverAddr, sizeof(receiverAddr));
	Netmusic::check_error(rc, "Error binding mini to receiverAddr");

	// output MIDI Event PGM Change
	snd_seq_event_t* event = new snd_seq_event_t;
	snd_seq_ev_clear(event);
	snd_seq_ev_set_pgmchange(event, 0, pgmChange);
	snd_seq_ev_set_subs(event);
	snd_seq_ev_set_direct(event);
	snd_seq_ev_set_source(event, portId);
	snd_seq_event_output_direct(receiverSeqHandler, event);
	snd_seq_free_event(event);
	std::cout << "Send PGMChange" << std::endl;

	
	// receive mini words
	while(1)
	{
		// read firstRead
		n =recvfrom(sd, firstRead, (sizeof(char)), 0,
			(struct sockaddr *) &senderAddr, (socklen_t *)sizeof(senderAddr));
		std::cout << DELIMITER << std::endl;
		std::cout << "RECEIVER: Received 1st read..." << std::endl;

		// extract size, k, noteOff
		Mini::extract_size(firstRead[0], &readSize);

		// secondRead
		secondRead = (char*)new char[readSize-1];
		n = recvfrom(sd, secondRead, ((readSize-1)*sizeof(char)), 0,
			(struct sockaddr *) &senderAddr, (socklen_t *)sizeof(senderAddr));
		std::cout << "RECEIVER: Received 2nd read... " << std::endl << std::endl;

		//extract value
		char* miniWord = (char*)new char[readSize];
		miniWord[0] = firstRead[0];
		for(int i=1;i<readSize;i++)
			miniWord[i] = secondRead[i-1];
		Mini::print_mem( miniWord, readSize);
		mpz_set_ui(chordCode,0);
		Mini::bin_2_mpz(miniWord, &k, chordCode, &noteOff);

		// convert to int array
		noteArray = new int[k];
		Mini::ksub_unrank(k, noteArray, chordCode);

		// print results
		gmp_printf(" rank:     %Zd\n", chordCode);
		Mini::print_array(noteArray, k, startKey);
		std::cout << " mpz_size: " << readSize << std::endl;
		std::cout << " k:        " << k << std::endl;
		std::cout << " noteOff:  " << noteOff << std::endl;
		//mpz_clear(chordCode);

		snd_seq_event_t* ev = new snd_seq_event_t;
		
		// read mini word noteOff, to check if it is a mini noteOn or off word
		if(noteOff==0)
		{
			// generate noteOns
			for(int i=0;i<k;i++)
			{
			  snd_seq_ev_clear(ev);
			  snd_seq_ev_set_noteon(ev, 0, noteArray[i]+startKey, velocity);
				snd_seq_ev_set_subs(ev);
				snd_seq_ev_set_direct(ev);
				snd_seq_ev_set_source(ev, portId);
				snd_seq_event_output_direct(receiverSeqHandler, ev);
				std::cout << " output NOTE ON: " << (int)(*ev).data.note.note
					<< std::endl;
				snd_seq_free_event(ev);
			}
		}
		// MINI noteOffword
		else if(noteOff==1)
		{
			// generate noteOffs
			for(int i=0;i<k;i++)
			{
			  snd_seq_ev_clear(ev);
			  snd_seq_ev_set_noteoff(ev, 0, noteArray[i]+startKey, velocity);
				snd_seq_ev_set_subs(ev);
				snd_seq_ev_set_direct(ev);
				snd_seq_ev_set_source(ev, portId);
				snd_seq_event_output_direct(receiverSeqHandler, ev);
				std::cout << " output NOTE OFF: " << (int)(*ev).data.note.note
					<< std::endl;
				snd_seq_free_event(ev);
			}
		}
	}
	
 	pthread_exit(NULL); 
}

//********************************************************************
/*
 * \brief syncThread function
 * \param ptr empty pointer
 *
 * Function listens for input on the sync_socket. Then the values
 * are set, according to the received values.
 *
 */
void *sync_function(void *ptr)
{
	// variables
	int* arrayPointer = new int[4];
	struct sockaddr_in senderAddr, receiverAddr;
	int sd, rc, n;

	// socket creation
	sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	Netmusic::check_error(sd, "Error creating snyc socket");
	// bind receiver
	receiverAddr.sin_family = AF_INET;
	receiverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	receiverAddr.sin_port = htons(syncPort);
	rc = bind(sd, (struct sockaddr *) &receiverAddr, sizeof(receiverAddr));
	Netmusic::check_error(rc, "Error binding sync socket to receiverAddr");
	// receive message
	std::cout << DELIMITER << std::endl;
	std::cout << "Waiting on port "<< syncPort <<" for sync data from the sender..." << std::endl;
	n = recvfrom(sd, arrayPointer, (4*sizeof(int)), 0,
		(struct sockaddr *) &senderAddr, (socklen_t *)sizeof(senderAddr));
	//	set parameters
	mini = arrayPointer[0];
	midiPort = arrayPointer[1];
	miniPort = arrayPointer[2];
	startKey = arrayPointer[3];
	std::cout << "Sync complete!" << std::endl;
	
	close(sd);
	
	// end thread
	pthread_exit(NULL);
}

//********************************************************************
/**
 * \brief main function
 * \param argc Number of given arguments
 * \param argv[] Array of arguments
 *
 * Main function handles the parameters, as well as the thread creation
 *
 */
int main(int argc, char** argv)
{
	signal(SIGINT, handler);

	int miniRet, midiRet, syncRet;
	
	// set arguments
	Netmusic::receiver_check_args(argc, argv, 
		&syncPort, 
		&pgmChange, &velocity);

	// set alsa
	receiverSeqHandler = Netmusic::receiver_open_seq(&portId);
	npfd = snd_seq_poll_descriptors_count(receiverSeqHandler, POLLOUT);
	pfd = (struct pollfd *)alloca(npfd * sizeof(struct pollfd));
	snd_seq_poll_descriptors(receiverSeqHandler, pfd, npfd, POLLOUT);
	
	// print welcome message
	Netmusic::receiver_print_welcome_msg();
	
	// create sync thread
	syncRet = pthread_create( &syncThread, NULL, sync_function, NULL);
	// wait for syncThread
	pthread_join(syncThread, NULL);
	
	// print parameters
	Netmusic::receiver_print_param(
		mini,
		midiPort, miniPort, syncPort,
		startKey, pgmChange, velocity);
	
	// create other threads
	if(mini==0)
		midiRet = pthread_create( &midiThread, NULL, midi_function, NULL);
	if(mini==1)
		miniRet = pthread_create( &miniThread, NULL, mini_function, NULL);
	
	if(mini==0)
		pthread_join( midiThread, NULL);
	if(mini==1)
		pthread_join( miniThread, NULL);

	return 0;
}

