/**
 * \file netmusic_sender.cpp
 * \brief Sender program code.
 * \author <a href="mailto:auto_pilot@gmx.at">Hoermann Klaus</a>
 * \author <a target="_blank" href="http://www.netmusic.at.tt">www.netmusic.at.tt</a>
 * \version 0.1
 * \date 18.03.2005
 * \date last modified 31.05.2005
 *
 */

/**
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <sys/time.h>
#include <netdb.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/msg.h>
#include <sys/ipc.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <queue.h>

#include "include/Netmusic.hh"
#include "include/Mini.hh"

/**
 * ALSA Sound Sequencer handler
 */
snd_seq_t* senderSeqHandler = NULL;
/**
 * Max connections for poll function
 */
int npfd;
/**
 * Structure for poll descriptor
 */
struct pollfd* pfd;
/**
 * Socket descriptor for Mini connections
 */
int sdMini;
/**
 * The remote Machine Socketaddress.
 */
struct sockaddr_in remoteReceiverAddrMini;
/**
 * The Mini On queue holding Midi On note values
 */
queue<int>miniOnQueue;
/**
 * The Mini Off queue holding Midi Off note values
 */
queue<int>miniOffQueue;
/**
 * Parameter mini (default 1)
 */
int mini = 1;
/**
 * Parameter synchronization port (default 3000)
 */
int syncPort = 29998;
/**
 * Parameter midi port (default 3001)
 */
int midiPort = 29999;
/**
 * Parameter mini port (default 3002)
 */
int miniPort = 30000;
/**
 * Parameter start key (default 48)
 */
int startKey = 36;
/**
 * The IP address of the receiving computer
 */
char* receiverIp = new char[15];
/**
 * Time interval in which the mini on and mini off queue get emptied.
 */
int threshold = 200;
/**
 * Used threads: MiniOn, MiniOff, Midi, Sync
 */
pthread_t miniOnThread, miniOffThread, midiThread, syncThread;
/**
 * Mutex variable to lock the miniOnQueue
 */
pthread_mutex_t mutexOn = PTHREAD_MUTEX_INITIALIZER;
/**
 * Mutex variable to lock the miniOffQueue
 */
pthread_mutex_t mutexOff = PTHREAD_MUTEX_INITIALIZER;
/**
 * Mutex variable to lock the socket for Mini com
 */
pthread_mutex_t mutexMini = PTHREAD_MUTEX_INITIALIZER;


//********************************************************************
/**
 * \brief Signal handler for STRG-C
 * \param signal value for the given signal
 */
void handler(int signal)
{
	if ( signal == SIGINT)
		_exit(0); // kills the process and all threads
}

//********************************************************************
/*
 * \brief midiThread function
 * \param ptr empty pointer
 * 
 * This function handles the Midi input from a external ALSA sequencer
 * input port. Depending on the mini parameter, the incoming events
 * are sent directly to the receiver, or the NoteOn and NoteOff 
 * messages, are put in the associated queue.
 *
 */
void *midi_function(void *ptr)
{
	// welcome msg
	std::cout << "SENDER: MIDI Thread started..." << std::endl;

	// variables
	struct sockaddr_in senderAddrMidi, remoteReceiverAddrMidi;
	struct hostent *h;
	int sdMidi, rcMidi;
	snd_seq_event_t* ev;

	h = gethostbyname(receiverIp);

	// set receiver
	remoteReceiverAddrMidi.sin_family = h->h_addrtype;
	remoteReceiverAddrMidi.sin_port = htons(midiPort);
	memcpy((char *)&remoteReceiverAddrMidi.sin_addr.s_addr,
		h->h_addr_list[0], h->h_length);
	sdMidi = socket(AF_INET, SOCK_DGRAM, 0); // create socket
	Netmusic::check_error(sdMidi, "Error creating midi socket");
	// set sender
	senderAddrMidi.sin_family = AF_INET;
	senderAddrMidi.sin_addr.s_addr = htonl(INADDR_ANY);
	senderAddrMidi.sin_port = htons(0);
	// bind socket to sender structure
	rcMidi = bind(sdMidi, (struct sockaddr *) &senderAddrMidi, sizeof(senderAddrMidi));
	Netmusic::check_error(rcMidi, "Error binding to senderAddr");

	// listening for midi input
	while(1)
	{
		if( poll(pfd, npfd, 100000) > 0)
		{
			do
			{
				snd_seq_event_input(senderSeqHandler, &ev);
				
				Netmusic::print_read_data(ev);

				// midi chosen --> SEND MIDI
				if(mini==0)
				{
					if( ev->type == SND_SEQ_EVENT_NOTEON ||
					    ev->type == SND_SEQ_EVENT_NOTEOFF ||
						 ev->type == SND_SEQ_EVENT_PGMCHANGE )
					{
						rcMidi = sendto(sdMidi, ev, (sizeof(snd_seq_event_t)), 0,
							(struct sockaddr *) &remoteReceiverAddrMidi, sizeof(remoteReceiverAddrMidi));
						Netmusic::check_error(rcMidi, "Error sending to remote machine");
					}
				}
				
				// mini chosen --> SEND MINI, send Notes into queue
				else if(mini==1)
				{
					switch (ev->type)
					{
						case SND_SEQ_EVENT_NOTEON:
							// put occured noteOn note value into noteOnQueue
							// check velocity
							if(ev->data.note.velocity != 0)
							{
								pthread_mutex_lock(&mutexOn);
								miniOnQueue.push(ev->data.note.note);
								pthread_mutex_unlock(&mutexOn);
							}
							// note off into 'note_off_queue'
							else
							{
								pthread_mutex_lock(&mutexOff);
								miniOffQueue.push(ev->data.note.note);
								pthread_mutex_unlock(&mutexOff);
							}

							break;
						
						case SND_SEQ_EVENT_NOTEOFF:
							// put occured noteOff note value into 'note_off_queue'
							pthread_mutex_lock(&mutexOff);
							miniOffQueue.push(ev->data.note.note);
							pthread_mutex_unlock(&mutexOff);
							break;
						
					}
				}
				
				snd_seq_free_event(ev);

			} while( snd_seq_event_input_pending(senderSeqHandler, 0) > 0);
		}
	}
	pthread_exit(NULL);
}

//********************************************************************
/*
 * \brief miniOnThread function
 * \param ptr empty pointer
 *
 * The mini_on_function is executed by the miniOnThread. Every <threshold>
 * the thread looks at the miniOnQueue, and if necessary removes the elements,
 * to calculate the appropriate miniOnWord. Then the mini word is sent to
 * the receiver.
 *
 * \sa mini_off_function()
 */
void *mini_on_function(void *ptr)
{
	// welcome msg
	std::cout << "SENDER: MINI ON Thread started..." << std::endl;

	// variables
	mpz_t miniOnWord;
	mpz_t tempWord1, tempWord2;
	char* binaryMiniOnWord;
	int* noteOnArray;
	int size, i=0, size_mem_mpz, localNoteOff=0, rcMini, tmp;

	// init mpz
	mpz_init(miniOnWord);
	mpz_init(tempWord1);
	mpz_init(tempWord2);
	mpz_set_ui(miniOnWord,0);

	// mini queue loop
	while(1)
	{
		usleep(threshold*1000); // wait threshold in microseconds

		pthread_mutex_lock(&mutexOn); // lock queue

		if(!miniOnQueue.empty()) // only if there is something in the queue
		{
			size = miniOnQueue.size(); // size of queue
			noteOnArray = new int[size]; // allocate memory for array

			while(!miniOnQueue.empty()) // as long as not empty
			{
				tmp = miniOnQueue.front(); // get top element
				noteOnArray[i] = tmp-startKey; // get note value relative to start key
				miniOnQueue.pop(); // pop top element
				i++; // next element
			}
			pthread_mutex_unlock(&mutexOn); // unlock queue


			// mini operations
			mpz_set_ui(miniOnWord,0);
			Mini::sort_array(noteOnArray, size);// sort array
			Mini::ksub_rank( size, noteOnArray, miniOnWord);// get miniOnWord
			mpz_set(tempWord1, miniOnWord); // set temp value
			mpz_set(tempWord2, miniOnWord); // set temp value
			size_mem_mpz = Mini::get_size_mem_mpz(tempWord1); // get bin_size of tempValue
			binaryMiniOnWord = (char*)new char[size_mem_mpz]; // allocate memory
			Mini::mpz_2_bin(tempWord2, size, localNoteOff, binaryMiniOnWord); // convert to bin

			// print
			std::cout << DELIMITER << std::endl;
			Mini::print_array(noteOnArray, size, startKey);// print array
			gmp_printf("\n rank: %Zd\n", miniOnWord);
			Mini::print_mem( binaryMiniOnWord, size_mem_mpz);
			std::cout << " mpz_size: " << size_mem_mpz << std::endl;
			std::cout << " k:        " << size << std::endl;
			std::cout << " noteOff:  " << localNoteOff << std::endl << std::endl;
			
			// send binary to receiver --> using Mini port
			pthread_mutex_lock(&mutexMini);
			std::cout << "SENDER: Sending MINI ON WORD 1st Read to receiver..." << std::endl;
			rcMini = sendto(sdMini, binaryMiniOnWord, (sizeof(char)), 0,
				(struct sockaddr *) &remoteReceiverAddrMini, sizeof(remoteReceiverAddrMini));
			Netmusic::check_error(rcMini, "Error sending 1st read on");
			binaryMiniOnWord++;
			std::cout << "SENDER: Sending MINI ON WORD 2nd Read to receiver..." << std::endl;
			rcMini = sendto(sdMini, binaryMiniOnWord, ((size_mem_mpz-1)*sizeof(char)), 0,
				(struct sockaddr *) &remoteReceiverAddrMini, sizeof(remoteReceiverAddrMini));
			Netmusic::check_error(rcMini, "Error sending 2nd read on");
			pthread_mutex_unlock(&mutexMini);

			i=0; // set i to 0 for new attempt
		}

		pthread_mutex_unlock(&mutexOn); // unlock queue if queue empty !!!!
	}
	close(sdMini);
}

//********************************************************************
/*
 * \brief miniOffThread function
 * \param ptr empty pointer
 *
 * The mini_off_function is executed by the miniOffThread. Every <threshold>
 * the thread looks at the miniOffQueue, and if necessary removes the elements,
 * to calculate the appropriate miniOffWord. Then the mini word is sent to
 * the receiver.
 *
 * \sa mini_on_function
 */
void *mini_off_function(void *ptr)
{
	// welcome msg
	std::cout << "SENDER: MINI OFF Thread started..." << std::endl;

	// variables
	mpz_t miniOffWord;
	mpz_t tempWord1, tempWord2;
	char* binaryMiniOffWord;
	int* noteOffArray;
	int size, i=0, size_mem_mpz, localNoteOff=1, rcMini, tmp;

	// init mpz
	mpz_init(miniOffWord);
	mpz_init(tempWord1);
	mpz_init(tempWord2);
	mpz_set_ui(miniOffWord,0);

	// mini queue loop
	while(1)
	{
		usleep(threshold*1000); // wait threshold in microseconds

		pthread_mutex_lock(&mutexOff); // lock queue

		if(!miniOffQueue.empty()) // only if there is something in the queue
		{
			size = miniOffQueue.size(); // size of queue
			noteOffArray = new int[size]; // allocate memory for array

			while(!miniOffQueue.empty()) // as long as not empty
			{
				tmp = miniOffQueue.front(); // get top element
				noteOffArray[i] = tmp-startKey; // get note value relative to start key
				miniOffQueue.pop(); // pop top element
				i++; // next element
			}
			pthread_mutex_unlock(&mutexOff); // unlock queue

			// mini operations
			mpz_set_ui(miniOffWord,0);
			Mini::sort_array(noteOffArray, size);// sort array
			Mini::ksub_rank( size, noteOffArray, miniOffWord);// get miniOnWord
			mpz_set(tempWord1, miniOffWord); // set temp value
			mpz_set(tempWord2, miniOffWord); // set temp value
			size_mem_mpz = Mini::get_size_mem_mpz(tempWord1); // get bin_size of tempValue
			binaryMiniOffWord = (char*)new char[size_mem_mpz]; // allocate memory
			Mini::mpz_2_bin(tempWord2, size, localNoteOff, binaryMiniOffWord); // convert to bin

			// print
			std::cout << DELIMITER << std::endl;
			Mini::print_array(noteOffArray, size, startKey);// print array
			gmp_printf("\n rank: %Zd\n", miniOffWord);
			Mini::print_mem( binaryMiniOffWord, size_mem_mpz);
			std::cout << " mpz_size: " << size_mem_mpz << std::endl;
			std::cout << " k:        " << size << std::endl;
			std::cout << " noteOff:  " << localNoteOff << std::endl << std::endl;

			// send binary to receiver --> using Mini port
			pthread_mutex_lock(&mutexMini);
			std::cout << "SENDER: Sending MINI OFF WORD 1st Read to receiver..." << std::endl;
			rcMini = sendto(sdMini, binaryMiniOffWord, (sizeof(char)), 0,
				(struct sockaddr *) &remoteReceiverAddrMini, sizeof(remoteReceiverAddrMini));
			Netmusic::check_error(rcMini, "Error sending 1st read off");
			binaryMiniOffWord++;
			std::cout << "SENDER: Sending MINI OFF WORD 2nd Read to receiver..." << std::endl;
			rcMini = sendto(sdMini, binaryMiniOffWord, ((size_mem_mpz-1)*sizeof(char)), 0,
				(struct sockaddr *) &remoteReceiverAddrMini, sizeof(remoteReceiverAddrMini));
			Netmusic::check_error(rcMini, "Error sending 2nd read off");
			pthread_mutex_unlock(&mutexMini);

			i=0; // set i to 0 for new attempt
		}

		pthread_mutex_unlock(&mutexOff); // unlock queue if queue empty !!!!
	}
	close(sdMini);
}

//********************************************************************
/*
 * \brief syncThread function
 * \param ptr empty pointer
 *
 * Method for the sync thread. It sends the synchronized parameters
 * in form of an array to the receiver.
 *
 */
void *sync_function(void *ptr)
{
	// variables
	struct sockaddr_in senderAddrSync, remoteReceiverAddr;
	struct hostent *h;
	int sd, rc;
	int* arrayPointer = new int[4];

	h = gethostbyname(receiverIp);

	// set receiver
	remoteReceiverAddr.sin_family = h->h_addrtype;
	remoteReceiverAddr.sin_port = htons(syncPort);
	memcpy((char *)&remoteReceiverAddr.sin_addr.s_addr, 
		h->h_addr_list[0], h->h_length);
	// create socket
	sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	Netmusic::check_error(sd, "Error creating sync socket");
	// set sender
	senderAddrSync.sin_family = AF_INET;
	senderAddrSync.sin_addr.s_addr = htonl(INADDR_ANY);
	senderAddrSync.sin_port = htons(0);
	// bind socket to sender
	rc = bind(sd, (struct sockaddr *) &senderAddrSync, sizeof(senderAddrSync));
	Netmusic::check_error(rc, "Error binding to senderAddr");

	// set array for sync
	arrayPointer[0] = mini;
	arrayPointer[1] = midiPort;
	arrayPointer[2] = miniPort;
	arrayPointer[3] = startKey;

	rc = sendto(sd, arrayPointer, (4*sizeof(int)), 0,
		(struct sockaddr *) &remoteReceiverAddr, sizeof(remoteReceiverAddr));
	Netmusic::check_error(rc, "Error sending sync data");

	close(sd);
	
	std::cout << DELIMITER << std::endl;
	std::cout << "Sending sync data to receiver..." << std::endl;
	std::cout << "Sync complete." << std::endl;

	pthread_exit(NULL);
}

//********************************************************************
/**
 * \brief main function
 * \param argc Number of given arguments.
 * \param argv[] Array of arguments.
 *
 * Main function handles the thread operations, as well as the
 * parameter settings.
 *
 */
int main(int argc, char** argv)
{
	signal(SIGINT, handler);

	// set arguments
	Netmusic::sender_check_args(argc, argv,
		&mini,
		&midiPort, &miniPort, &syncPort,
		&startKey, &threshold,
		receiverIp);

	// set global mini UDP connection
	int rcMini;
	struct sockaddr_in senderAddrMini;
	struct hostent *h;
	h = gethostbyname(receiverIp);
	// set receiver
	remoteReceiverAddrMini.sin_family = h->h_addrtype;
	remoteReceiverAddrMini.sin_port = htons(miniPort);
	memcpy((char *)&remoteReceiverAddrMini.sin_addr.s_addr,
		h->h_addr_list[0], h->h_length);
	// create socket
	sdMini = socket(AF_INET, SOCK_DGRAM, 0);
	Netmusic::check_error(sdMini, "Error creating mini socket");
	// set sender
	senderAddrMini.sin_family = AF_INET;
	senderAddrMini.sin_addr.s_addr = htonl(INADDR_ANY);
	senderAddrMini.sin_port = htons(0);
	// bind socket with sender
	rcMini = bind(sdMini, (struct sockaddr *) &senderAddrMini, sizeof(senderAddrMini));
	Netmusic::check_error(rcMini, "Error binding mini socket");

	// set alsa
	senderSeqHandler = Netmusic::sender_open_seq();
	npfd = snd_seq_poll_descriptors_count(senderSeqHandler, POLLIN);
	pfd = (struct pollfd *)alloca(npfd * sizeof(struct pollfd));
	snd_seq_poll_descriptors(senderSeqHandler, pfd, npfd, POLLIN);

	// print welcome message
	Netmusic::sender_print_welcome_msg();

	int miniOnRet, miniOffRet, midiRet, syncRet; // thread return values

	// create sync thread
	syncRet = pthread_create( &syncThread, NULL, sync_function, NULL);
	// wait for syncThread termination
	pthread_join( syncThread, NULL);
			
	// print parameters
	Netmusic::sender_print_param(
		mini,
		midiPort, miniPort, syncPort,
		startKey, threshold,
		receiverIp);
		
	// create midiThread always
	midiRet = pthread_create( &midiThread, NULL, midi_function, NULL);
	// create MiniOnThread only if mini is chosen
	if(mini==1)
	{
		miniOnRet = pthread_create( &miniOnThread, NULL, mini_on_function, NULL);
		miniOffRet = pthread_create( &miniOffThread, NULL, mini_off_function, NULL);
	}

	// wait for midi thread to finish
	pthread_join( midiThread, NULL);
	// wait for miniOnthread'
	if(mini==1)
	{
		pthread_join( miniOnThread, NULL);
		pthread_join( miniOffThread, NULL);
	}
	
	return 0;
}
